
; SetSystem.s (07-Feb-2009)
; robert.leffmann@gmail.com
;
; Allows the user to query and temporarily disable the system with the option
; to degrade it by changing the VBR, cache and video settings to make it
; closer resemble a specific system.
;
; Fix:
;  Allocate audio and timers in an OS friendly way


                    include hardware.i
                    include macros.i
                    include exec/execbase.i
                    include exec/libraries.i
                    include graphics/gfxbase.i
                    include lvo/exec.i
                    include lvo/graphics.i
                    include lvo/intuition.i

                    public SetSystem, RestoreSystem

;------------------------------------------------------------------------------
                    section code, code

; Arguments
;
; D0 Kickstart requirement    D1 CPU requirement    D2 Chipset requirement
;
; -1 = any                    -1 = any              -1 = any
; 33 = 1.2                     0 = 68000             0 = OCS
; 34 = 1.3                     1 = 68010             1 = ECS
; 36 = 2.0                     2 = 68020             2 = AGA
; 37 = 2.04                    3 = 68030
; 39 = 3.0                     4 = 68040            D3 Operation
; 40 = 3.1                     6 = 68060
;                             1x = 68881             0 = Read metrics
;                             2x = 68882             1 = Disable system
;                                                    2 = Disable and degrade
; Return values
;
; D0-D2 Actual system metrics as above
;    D3 Pointer to vector base
;    D4 0 = Requirements not met

SetSystem           lea       variables, a2
                    lea       requestedsystem, a3
                    lea       actualsystem, a4
                    clr.w     (cachechanged, a2)

                    bsr       ReadSystem          ; Read system values
                    or.b      d3, d4
                    beq       .return             ; Return if we can not match
                                                  ; the required system
                    pushw     d3
                    bsr       Disable             ; Temporarily disable the
                    popw      d3                  ; system, or return if there
                    tst.b     d4                  ; was an error
                    beq       .return

                    moveq     #1, d4
                    cmp.b     #2, d3
                    bne       .return
                    bsr       Degrade             ; Degrade the system

.return             moveq     #0, d0
                    moveq     #0, d2
                    move.w    (kick, a4), d0      ; Return the actual system
                    moveq     #10, d1             ; values to the user
                    mulu.w    (fpu, a4), d1
                    add.w     (cpu, a4), d1
                    move.w    (chip, a4), d2
                    move.l    (newvbr, a2), d3
                    rts



                    ; Disable the system in an OS friendly way so that we can
                    ; poke the hardware and still resume normal operation after
                    ; our program has finished

Disable             lea       .gfxlib, a1
                    execall   OldOpenLibrary
                    move.l    d0, _GfxBase

                    lea       .intlib, a1
                    syscall   OldOpenLibrary
                    move.l    d0, _IntuitionBase

                    sub.l     a1, a1
                    syscall   FindTask
                    moveq     #127, d0
                    syscall   SetTaskPri

                  ;  bsr       .gettimers
                  ;  bsr       .getaudio
                    bsr       .getvideo

                    moveq     #1, d4
                    rts

.getvideo           gfxcall   OwnBlitter
                    syscall   WaitBlit
                    sub.l     a1, a1
                    syscall   LoadView            ; Set the internal display
                    syscall   WaitTOF             ; registers in a default
                    syscall   WaitTOF             ; state and blank the screen

                    move.w    intenar, (oldintena, a2)
                    move.w    intreqr, (oldintreq, a2)
                    move.w    dmaconr, (olddmacon, a2)
                    move.w    #$7fff, intena
                    move.w    #$7fff, intreq      ; Disable all DMA
                    move.w    #$7ff, dmacon       ; and interrupts
                    rts

.gfxlib             dc.b "graphics.library", 0
.intlib             dc.b "intuition.library", 0
                    even



                    ; Degrade the video and CPU settings to make the system
                    ; behave a bit closer to what we have requested
                    
Degrade             cmp.w     #2, (chip, a3)
                    bhs       .noocs

                    move.w    #0, fmode           ; Set video settings to
                    move.w    #$200, bplcon0      ; OCS default values
                    move.w    #0, bplcon2    
                    move.w    #$c40, bplcon3
                    move.w    #$11, bplcon4

.noocs              cmp.w     #-1, (cpu, a3)      ; Return if we have a 68000
                    beq       .done               ; or do not want to degrade
                    tst.w     (cpu, a4)
                    beq       .done
                    tst.w     (cpu, a3)
                    bne       .vbrok
                    
                    move.l    (oldvbr, a2), a0    ; If we are degrading from a
                    sub.l     a1, a1              ; higher CPU down to 68000
                    moveq     #64-1, d0           ; then we need to copy the
.copyvecs           move.l    (a0)+, (a1)+        ; vector base to address 0
                    dbf       d0, .copyvecs       ; and point the VBR there

                    moveq     #0, d7
                    move.l    d7, (newvbr, a2)
                    lea       GetVBR, a5
                    execall   Supervisor

.vbrok              bsr       DegradeCache

.done               rts



                    ; Disable cache features to make the processor closer
                    ; resemble the model we have requested

DegradeCache        move.w    (cpu, a4), d3
                    move.w    (cpu, a3), d2       ; If we did not ask for a
                    cmp.w     d3, d2              ; lower spec CPU then we
                    bge       .done               ; do not need to degrade

                    move.l    4, a6
                    cmp.w     #2, d3              ; See what CPU we have and
                    beq       Deg020              ; jump to the right routine
                    cmp.w     #3, d3
                    beq       Deg030
                    cmp.w     #4, d3
                    beq       Deg040
                    cmp.w     #6, d3
                    beq       Deg060              ; Lower than 68020 means we
.done               rts                           ; have no cache


Deg060              cmp.w     #4, d2              ; Degrade the 68060
                    bne       .not040
                    lea       .to040, a5
.not040             cmp.w     #3, d2
                    bne       .not030
                    lea       .to030, a5
.not030             cmp.w     #2, d2
                    bne       .not020
                    lea       .to020, a5
.not020             cmp.w     #1, d2
                    bgt       .degrade
                    lea       .off, a5

.degrade            syscall   Supervisor
                    move.l    d7, (oldcache, a2)
                    st        (cachechanged, a2)
                    rts

.off                movec     cacr, d7
                    move.l    #1<<22, d0
                    movec     d0, cacr
                    cpusha    bc
                    rte
.to040
.to030              movec     cacr, d7
                    move.l    #(1<<31)|(1<<22)|(1<<15), d0
                    movec     d0, cacr
                    rte

.to020              movec     cacr, d7
                    move.l    #(1<<22)|(1<<15), d0
                    movec     d0, cacr
                    cpusha    bc
                    rte

                    
Deg040              cmp.w     #3, d2              ; Degrade the 68040
                    bne       .not030
                    lea       .to030, a5
.not030             cmp.w     #2, d2
                    bne       .not020
                    lea       .to020, a5
.not020             cmp.w     #1, d2
                    bgt       .degrade
                    lea       .off, a5

.degrade            syscall   Supervisor
                    move.l    d7, (oldcache, a2)
                    st        (cachechanged, a2)
                    rts

.off                movec     cacr, d7
                    moveq     #0, d0
                    movec     d0, cacr
                    cpusha    bc
                    rte

.to030              movec     cacr, d7
                    rte
                    
.to020              movec     cacr, d7
                    move.l    #1<<15, d0
                    movec     d0, cacr
                    cpusha    bc
                    rte

                    
Deg030              cmp.w     #2, d2              ; Degrade the 68030
                    bne       .not020
                    lea       .to020, a5
.not020             cmp.w     #1, d2
                    bgt       .degrade
                    lea       .off, a5

.degrade            syscall   Supervisor
                    move.l    d7, (oldcache, a2)
                    st        (cachechanged, a2)
                    rts

.off                movec     cacr, d7
                    move.l    #(1<<11)|(1<<3), d0
                    movec     d0, cacr
                    rte

.to020              movec     cacr, d7
                    move.l    #(1<<11)|(1<<3) | 1, d0
                    movec     d0, cacr
                    rte


Deg020              lea       .off, a5            ; Degrade the 68020
                    syscall   Supervisor
                    move.l    d7, (oldcache, a2)
                    st        (cachechanged, a2)
                    rts

.off                movec     cacr, d7
                    moveq     #1<<3, d0
                    movec     d0, cacr
                    rte



                    ; Get or set the VBR

GetVBR              cmp.b     #1, d7
                    bne       .set
                    movec     vbr, d7
                    rte
.set                movec     d7, vbr
                    rte



                    ; Read all the system specific values and see if they match
                    ; the requested system

ReadSystem          move.w    d0, (kick, a3)
                    divu.w    #10, d1
                    move.w    d1, (fpu, a3)
                    swap      d1
                    move.w    d1, (cpu, a3)
                    move.w    d2, (chip, a3)

                    move.l    4, a6
                    move.w    (LIB_VERSION, a6), (kick, a4)

                    move.w    (AttnFlags, a6), d5
                    moveq     #0, d6
                    moveq     #0, d7
                    moveq     #4-1, d4            ; See what CPU and FPU we
.getcpu             lsr.b     #1, d5              ; have installed
                    addx.b    d7, d6
                    dbf       d4, .getcpu
                    moveq     #0, d4
                    lsr.b     #1, d5
                    addx.b    d7, d4
                    lsr.b     #1, d5
                    addx.b    d7, d4
                    move.w    d4, (fpu, a4)
                    bclr      #0, d5
                    add.b     d5, d6
                    move.w    d6, (cpu, a4)

                    moveq     #0, d7              ; If we have a 68000 then the
                    tst.w     d6                  ; vector base is at address 0
                    beq       .novbr

                    moveq     #1, d7
                    lea       GetVBR, a5          ; Otherwise we read the VBR
                    syscall   Supervisor

.novbr              move.l    d7, (oldvbr, a2)
                    move.l    d7, (newvbr, a2)

                    moveq     #0, d7
                    move.w    lisaid, d4          ; Check for OCS/ECS/AGA
                    cmp.b     #$fc, d4
                    bne       .noecs
                    moveq     #1, d7
.noecs              cmp.b     #$f8, d4
                    bne       .noaga
                    moveq     #2, d7
.noaga              move.w    d7, (chip, a4)

                    moveq     #0, d4
                    move.w    (kick, a4), d0        ; See if our requested values
                    cmp.w     (kick, a3), d0        ; match the system values
                    blt       .done
                    move.w    (cpu, a4), d0
                    cmp.w     (cpu, a3), d0
                    blt       .done
                    move.w    (fpu, a4), d0
                    cmp.w     (fpu, a3), d0
                    blt       .done
                    move.w    (chip, a4), d0
                    cmp.w     (chip, a3), d0
                    blt       .done

                    moveq     #1, d4
.done               rts



                    ; Restore and resume normal system operation

RestoreSystem       push      d0

                    lea       variables, a2
                    lea       actualsystem, a4

                    gfxcall   WaitBlit            ; Wait for the blitter to
                                                  ; finish if it is running

                    move.w    #$7fff, intena      ; Disable all DMA and
                    move.w    #$7fff, intreq      ; interrupts
                    move.w    #$7ff, dmacon

                    tst.w     (cpu, a4)           ; If we have a 68000 then
                    beq       .cpuok              ; there is no VBR or cache
                                                  ; to restore

                    move.l    (oldvbr, a2), d7    ; Otherwise we set the VBR
                    lea       GetVBR, a5          ; back to its old value
                    execall   Supervisor

                    tst.w     (cachechanged, a2)  ; Restore cache settings if
                    beq       .cpuok              ; we degraded them earlier

                    lea       .restorecache, a5
                    syscall   Supervisor

.cpuok              move.l    _GfxBase, a0
                    move.l    (gb_copinit, a0), cop1lc
                    move.w    d0, copjmp1
            
                    move.w    (oldintena, a2), d0
                    move.w    (oldintreq, a2), d1 ; Restore previous DMA and
                    move.w    (olddmacon, a2), d2 ; interrupt settings
                    or.w      #int_set, d0
                    or.w      #int_set, d1
                    or.w      #dma_set, d2
                    move.w    d0, intena
                    move.w    d1, intreq
                    move.w    d2, dmacon

                    intcall   RemakeDisplay       ; Restore the old view
                    gfxcall   WaitTOF             ; Let the display changes
                    syscall   WaitTOF             ; take effect, and release
                    syscall   DisownBlitter       ; ownership of the blitter

                    move.l    a6, a1              ; Close libraries and exit
                    execall   CloseLibrary
                    move.l    _IntuitionBase, a1
                    syscall   CloseLibrary

                    pop       d0
                    rts

.restorecache       move.l    (oldcache, a2), d0
                    movec     d0, cacr
                    rte


;------------------------------------------------------------------------------
                    section bss, bss

                    clrso
oldintena           so.w 1
oldintreq           so.w 1
olddmacon           so.w 1
oldvbr              so.l 1
newvbr              so.l 1
cachechanged        so.w 1
oldcache            so.l 1
sizeof_var          so.x 0

                    clrso
cpu                 so.w 1
fpu                 so.w 1
chip                so.w 1
kick                so.w 1
sizeof_sys          so.x 0

_GfxBase            ds.l 1
_IntuitionBase      ds.l 1

variables           ds.b sizeof_var
                    even
actualsystem        ds.b sizeof_sys
                    even
requestedsystem     ds.b sizeof_sys
